import copy
import json
import urllib
from common import *
from datetime import datetime
import requests
import time
import httplib
import traceback


def event_object(event):
    print(event)
    bucket = event['Records'][0]['s3']['bucket']['name']
    key = urllib.unquote_plus(event['Records'][0]['s3']['object']['key'].encode('utf8'))
    file_size = int(event['Records'][0]['s3']['object']['size'])
    if (not bucket) or (not key):
        print("Unable to retrieve object from event.\n%s" % event)
        raise Exception("Unable to retrieve object from event.")
    if file_size >> 20 < 25:
        return s3.Object(bucket, key), True
    else:
        return s3.Object(bucket, key), False


def change_acl(bucket, key, acl='private'):
    object_acl = s3.ObjectAcl(bucket, key)
    response = object_acl.put(ACL=acl)
    return response


def download_s3_object(s3_object, local_prefix):
    local_path = "%s/%s/%s" % (local_prefix, s3_object.bucket_name, s3_object.key)
    create_dir(os.path.dirname(local_path))
    s3_object.download_file(local_path)
    return local_path


def set_vv_tags(s3_object, sha1, verdict, file_status):
    curr_tags = s3_client.get_object_tagging(Bucket=s3_object.bucket_name, Key=s3_object.key)["TagSet"]
    new_tags = copy.copy(curr_tags)
    for tag in curr_tags:
        if tag["Key"] in [VV_STATUS_TAG, VV_TIMESTAMP_TAG, VV_SHA1_TAG, VV_VERDICT_TAG]:
            new_tags.remove(tag)

    new_tags.append({"Key": VV_SHA1_TAG, "Value": str(sha1)})
    new_tags.append({"Key": VV_VERDICT_TAG, "Value": str(verdict)})
    new_tags.append({"Key": VV_STATUS_TAG, "Value": str(file_status)})
    new_tags.append({"Key": VV_TIMESTAMP_TAG, "Value": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S UTC")})
    s3_client.put_object_tagging(
        Bucket=s3_object.bucket_name,
        Key=s3_object.key,
        Tagging={"TagSet": new_tags}
    )


def calculate_sha1(file_full_path):
    import hashlib
    BUF_SIZE = 65536
    try:
        sha1 = hashlib.sha1()
        with open(file_full_path, 'rb') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                sha1.update(data)
        return sha1.hexdigest()
    except Exception:
        print("Error in calculate_sha1 file_path : %s " % file_full_path)
        return None


def get_api_key_header():
    headers = {'X-Api-Key': VV_API_KEY}
    return headers


def vv_file_query_api(sha1):
    json_response = None
    request_url = VALKYRIE_FILE_QUERY_API + sha1
    response = requests.request('GET', request_url, headers=get_api_key_header())
    if response.status_code == httplib.OK:
        json_response = json.loads(response.text)
    return json_response


def vv_file_upload_api(file_path):
    json_response = None
    if os.path.getsize(file_path) >> 20 < 25:
        with open(file_path, 'rb') as f:
            file_data = {
                "file_data": f
            }
            response = requests.post(VALKYRIE_FILE_UPLOAD_API, files=file_data, timeout=(30, 30), headers=get_api_key_header())
            if response.status_code == httplib.OK:
                json_response = json.loads(response.text)
    return json_response


def wait_for_analysis(sha1):
    verdict = None
    file_status = "UPLOAD_TIMEOUT"
    time_passed = 0

    while verdict is None and time_passed <= UPLOAD_WAIT_TIMEOUT:
        query_response = vv_file_query_api(sha1)
        if query_response:
            query_rcode = query_response.get("return_code")
            if query_rcode == VVReturnCodes.SUCCESS:
                # File exists on VV
                verdict = query_response.get("verdict")
                file_status = "UPLOAD_SUCCESS"
                break
            elif query_rcode in VVReturnCodes.REQUEST_LIMIT_REACHED:
                # Request limit reached abort process
                file_status = "QUERY_LIMIT_REACHED"
                break
        # Sleep before checking again
        time.sleep(UPLOAD_CHECK_INTERVAL)
        time_passed += UPLOAD_CHECK_INTERVAL

    return verdict, file_status


def start_scan_process(sha1, file_path):
    query_response = vv_file_query_api(sha1)
    verdict = None
    file_status = ""

    if query_response:
        query_rcode = query_response.get("return_code")
        if query_rcode == VVReturnCodes.SUCCESS:
            # File exists on VV
            verdict = query_response.get("verdict")
            file_status = "QUERY_SUCCESS"
        elif query_rcode == VVReturnCodes.FILE_NOT_EXISTS:
            # File does not exists on VV so upload file for scan
            upload_response = vv_file_upload_api(file_path)
            if upload_response:
                upload_rcode = upload_response.get("return_code")
                if upload_rcode == VVReturnCodes.SUCCESS:
                    # Upload success wait for completion
                    verdict, file_status = wait_for_analysis(sha1)
                elif upload_rcode in VVReturnCodes.REQUEST_LIMIT_REACHED:
                    # Defined request limit is reached for given api key, so skip file
                    file_status = "UPLOAD_LIMIT_REACHED"
                else:
                    # Unexpected response returned, skip file
                    file_status = "UPLOAD_FAIL_%s" % upload_rcode
            else:
                pass
        elif query_rcode in VVReturnCodes.REQUEST_LIMIT_REACHED:
            # Defined request limit is reached for given api key, so skip file
            file_status = "QUERY_LIMIT_REACHED"
        else:
            # Unexpected response returned, skip file
            file_status = "QUERY_FAIL_%s" % query_rcode

    return verdict, file_status


def lambda_handler(event, context):
    start_time = datetime.utcnow()
    print("VV-Check starting at %s\n" % (start_time.strftime("%Y/%m/%d %H:%M:%S UTC")))
    s3_object, valid_size = event_object(event)
    # s3_object = s3.Object("vv-analysis-bucket", "1uBJEhua.exe")
    if VV_CHANGE_PERMISSION:
        try:
            print("Change permission is true , changing permission to private")
            response = change_acl(s3_object.bucket_name, s3_object.key, 'private')
            if response['ResponseMetadata']['HTTPStatusCode'] == httplib.OK:
                print("Changed permission to private successfully")
        except:
            print("Exception in setting permissions to private err %s" % traceback.format_exc())
    if valid_size:
        file_path = download_s3_object(s3_object, "/tmp")
        # file_path = "/tmp/vv-analysis-bucket/1uBJEhua.exe"
        sha1 = calculate_sha1(file_path)
        verdict, file_status = start_scan_process(sha1, file_path)
        verdict = VALKYRIE_VERDICT_MAP.get(verdict, "UNKNOWN")
        print("Scan of s3://%s resulted - SHA1: %s VERDICT: %s STATUS: %s\n" % (os.path.join(s3_object.bucket_name, s3_object.key), sha1, verdict, file_status))
        set_vv_tags(s3_object, sha1, verdict, file_status)

        if VV_CHANGE_PERMISSION:
            try:
                if VV_ALLOWED_VERDICTS and len(VV_ALLOWED_VERDICTS)>0:
                    allowed_verdict_text = VV_ALLOWED_VERDICTS.upper()
                else:
                    allowed_verdict_text = "CLEAN,UNDETECTED"
                if verdict in allowed_verdict_text:
                    print("Change permission is true , changing permission to public-read")
                    response = change_acl(s3_object.bucket_name, s3_object.key, 'public-read')
                    if response['ResponseMetadata']['HTTPStatusCode'] == httplib.OK:
                        print("Changed permission to public-read successfully")
                else:
                    print("File verdict is not in allowed verdicts file verdict:%s allowed verdicts:%s" % (verdict, VV_ALLOWED_VERDICTS))
            except:
                print("Exception in reverting permissions err %s" % traceback.format_exc())

        try:
            os.remove(file_path)
        except OSError:
            pass
        print("VV-Check finished at %s\n" % datetime.utcnow().strftime("%Y/%m/%d %H:%M:%S UTC"))
    else:
        print("Scan of s3://%s resulted -Exceeded Maximum size limit 25MB skipping file \n" % (os.path.join(s3_object.bucket_name, s3_object.key)))


if __name__ == '__main__':
    lambda_handler("event", None)
